#!/usr/bin/env python3
from __future__ import with_statement, print_function
from contextlib import closing
from cgi import escape as H
import sys
import re
import sqlite3
try:
    import html5lib
except ImportError:
    print("html5lib not installed; try sudo apt-get install python3-html5lib")
    exit(1)
from os.path import join as pathjoin
from xml.parsers.expat import ExpatError

if True:
    imagesFolder = u'../wiki-images'
    outFolder = u'wikipages'
    dbFilename = "nesdevwiki.sqlite"
    urlPrefix = u"http://wiki.nesdev.com/w/index.php/"
    altUrlPrefix = u"http://wiki.nesdev.com/w/index.php?"
    skinsPrefix = u"/w/skins/"
    imgRE = re.compile(u"/w/images/([0-9a-f])/([0-9a-f]{2})/(.*)")
    thumbRE = re.compile(u"/w/images/thumb/([0-9a-f])/([0-9a-f]{2})/([^/]*)/([0-9]+)px-")
    deletedArticles = set([
        u"LJ65"
    ])
else:
    imagesFolder = u'../p8wiki-images'
    outFolder = u'wikipages'
    dbFilename = "p8wiki.sqlite"
    urlPrefix = u"//pineight.com/mw/index.php?title="
    altUrlPrefix = urlPrefix
    skinsPrefix = u"/mw/skins/"
    imgRE = re.compile(u"/mw/images/([0-9a-f])/([0-9a-f]{2})/(.*)")
    thumbRE = re.compile(u"/mw/images/thumb/([0-9a-f])/([0-9a-f]{2})/([^/]*)/([0-9]+)px-")
    deletedArticles = set([
        u"LJ65"
    ])

nonalnumRE = re.compile("[^0-9a-zA-Z]")
redirRE = re.compile(r"#REDIRECT\s*\[\[(.*?)\]\]", re.IGNORECASE)
def title_to_filename(title):
    return nonalnumRE.sub(u'_', title)+u'.xhtml'

def link_to_struck(el, className='', title=''):
    p = el.parentNode
##    print(p.toxml())
    s = el.ownerDocument.createElement(u's')
    if className:
        s.setAttribute(u'class', className)
    if title:
        s.setAttribute(u'title', title)
    for kid in el.childNodes:
        el.removeChild(kid)
        s.appendChild(kid)
    p.insertBefore(s, el)
    p.removeChild(el)
##    print(p.toxml())

def get_all_redirects(db):
    stmt = """
SELECT pg.title, ln.to_title
FROM links AS ln
INNER JOIN pages AS pg ON pg.pageid = ln.from_pageid
WHERE manner='redirect'
"""
    with closing(db.cursor()) as c:
        c.execute(stmt)
        return dict(c.fetchall())

def fix_internal_link(el, redirects):
    from urllib.parse import unquote as pct_unquote
    from cgi import parse_qsl

    href = el.getAttribute(u'href')
    sfrag = href.split(u'#', 1)
    fragment = u'#' + sfrag[1] if len(sfrag) > 1 else ''
    linktitle = sfrag[0]
    if linktitle.startswith(urlPrefix):
        (pfx, linktitle) = (urlPrefix, linktitle[len(urlPrefix):])
    else:
        (pfx, linktitle) = (altUrlPrefix, linktitle[len(altUrlPrefix):])
    sfrag = (None, linktitle) if pfx.endswith('?') else linktitle.split('?', 1)
    cgiargs = dict(parse_qsl(sfrag[1])) if len(sfrag) > 1 else {}
    try:
        linktitle = cgiargs['title']
    except KeyError:
        linktitle = pct_unquote(sfrag[0])
    linktitle = linktitle.replace('_', ' ')

    # These links should be struck out
    if (linktitle.startswith('Special:')
        or u'diff' in cgiargs
        or u'oldid' in cgiargs
        or u'redlink' in cgiargs
        or cgiargs.get(u'action') in (u'history', u'edit')):
        link_to_struck(el)
        return
    elif cgiargs:
        print(linktitle, cgiargs)
    linktitle = redirects.get(linktitle, linktitle)
    el.setAttribute(u'href', title_to_filename(linktitle) + fragment)

def process_html(title, h, redirects):
    from xml.dom import minidom

    htitle = H(title)
    h = "".join((
        """<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
  "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml"><head><title>""",
        htitle,
        """</title>
<meta name="viewport" content="width=display-width" />
<link rel="stylesheet" type="text/css" href="w.css" />
<script type="text/javascript" src="w.js"></script>
</head><body><h1>""",
        htitle,
        """</h1><div class="article">
""",
        h,
        """</div></body></html>"""
    ))
    doc = minidom.parseString(h)
    body = doc.getElementsByTagName(u'body')[0]

    # Remove section edit links
    spans = [(el, set(el.getAttribute('class').split()))
             for el in doc.getElementsByTagName(u'span')]
    for el, classNames in spans:
        if 'editsection' in classNames:
            el.parentNode.removeChild(el)
            continue

    # Convert links
    banned_links = []
    for el in doc.getElementsByTagName(u'a'):
        href = el.getAttribute(u'href')
        if href.startswith((u'http://nesdev.parodius.com', u'http://pinocchio.jk0.org')):
            banned_links.append(href)
        try:
            urlPrefix
        except NameError:
            print(href, file=sys.stderr)
            raise ValueError("cannot be substituted")
        if href.startswith((urlPrefix, altUrlPrefix)):
            fix_internal_link(el, redirects)

    for el in doc.getElementsByTagName(u'img'):
        src = el.getAttribute(u'src')
        try:
            imgRE
        except NameError:
            print(src, file=sys.stderr)
            raise ValueError("cannot be substituted")
        m = imgRE.match(src) or thumbRE.match(src)
        if m:
            src = ''.join((imagesFolder, u'/', m.group(3)))
            el.setAttribute(u'src', src)
        elif src.startswith(skinsPrefix):
            # most of these are /mw/skins/common/images/blah-blah.png
            # I'll handle these by referencing images on commons.wikimedia.org
            src = "https://commons.wikimedia.org/w/skins/" + src[len(skinsPrefix):]
            el.setAttribute(u'src', src)
    xml = doc.toxml().encode('utf-8')
    doc.unlink()
    if banned_links:
        print("%s: %d defunct links" % (title, len(banned_links)), file=sys.stderr)
        print("\n".join(sorted(banned_links)), file=sys.stderr)
    return xml


def get_all_pages(db):
    s = """
SELECT pageid, title,
       CASE WHEN markup LIKE '#REDIRECT%' THEN markup ELSE html END AS txt,
       CASE WHEN markup LIKE '#REDIRECT%' THEN 1 ELSE 0 END AS isRedirect
FROM pages
"""
    with closing(db.cursor()) as c:
        c.execute(s)
        rows = c.fetchall()

    catlists = dict((row[0], set()) for row in rows)
    s = """
SELECT from_pageid, to_title
FROM links
WHERE manner = 'cat'
"""
    with closing(db.cursor()) as c:
        c.execute(s)
        cats = c.fetchall()
    for from_pageid, to_title in cats:
        try:
            catlists[from_pageid].add(to_title)
        except KeyError:
            print("nonexistent pageid %s is in category %s"
                  % (from_pageid, to_title))
    rows = [(row + (catlists[row[0]],)) for row in rows]
    return rows

def main():
    with closing(sqlite3.connect(dbFilename)) as db:
        rows = get_all_pages(db)
        redirects = get_all_redirects(db)

    print("%d pages" % len(rows))
    outfiles = []
    while rows:
        (pageid, title, xml, isRedirect, categories) = rows.pop()
##        if title in deletedArticles:
##            continue
        filename = title_to_filename(title)
        if isRedirect:
            m = redirRE.match(xml)
            if not m:
                print("Bad redirect at", title, file=sys.stderr)
                continue
            redirtarget = m.group(1)
            xml = ('See <a href="%s">%s</a>.'
                   % (H(title_to_filename(redirtarget)),
                      H(redirtarget)))

        if categories:
            xml = ''.join([
                xml, '<p class="categories">Categories: ',
                ', '.join('<a href="%s">%s</a>'
                           % (H(title_to_filename(cat)), H(cat.split(':')[-1]))
                          for cat in categories),
                '</p>'])
        try:
            xml = process_html(title, xml, redirects)
        except ExpatError as e:
            print("\nExpatError in %s:" % title, file=sys.stderr)
            with open("%s_mismatched.xml" % title, "wb") as outfp:
                outfp.writelines([
                    '<html><body>', xml.encode('utf-8'), '</body></html>'
                ])
            lines_before_article = 5  # html head, article title, etc.
            lineno = e.lineno - lines_before_article
            lines = xml.split('\n')[lineno - 3:lineno + 2]
            print("\n".join(lines), file=sys.stderr)
            print("%s\nminus %d" % (e, lines_before_article), file=sys.stderr)
        except Exception as e:
            print("\nError in %s:\n%s" % (title, e), file=sys.stderr)
            from traceback import print_exc
            print_exc()
        else:
            outfiles.append((filename, title, isRedirect, xml))

    # Windows and OS X use case-insensitive file systems by default
    # so don't write files that overlap
    articles_lc = set()
    redirs_lc = set()
    for (filename, title, isRedirect, xml) in outfiles:
        fnlower = filename.lower()
        lc = redirs_lc if isRedirect else articles_lc
        if fnlower in lc:
            print("%s case collision: %s"
                  % ("redirect" if isRedirect else "article", fnlower))
        else:
            lc.add(fnlower)

    # Now actually write them to the folder
    lc_skipped = set()
    while outfiles:
        (filename, title, isRedirect, xml) = outfiles.pop()

        # Don't write redirects whose name overlaps an article
        if isRedirect and filename.lower() in articles_lc:
            lc_skipped.add(title)
            continue
        with open(pathjoin(outFolder, filename), 'wb') as outfp:
            outfp.write(xml)
    if lc_skipped:
        print("Skipped these redirects for title case collision:")
        print("\n".join(sorted(lc_skipped)))


if __name__=='__main__':
    main()
